"""
Taken from: https://github.com/jesbu1/openvla/blob/main/prismatic/vla/datasets/rlds/oxe/configs.py
configs.py

Modified some configs to get rid of bad image obs keys and add bridge_V2 and such.

Defines per-dataset configuration (kwargs) for each dataset in Open-X Embodiment.

Configuration adopts the following structure:
    image_obs_keys:
        primary: primary external RGB
        secondary: secondary external RGB
        wrist: wrist RGB

    depth_obs_keys:
        primary: primary external depth
        secondary: secondary external depth
        wrist: wrist depth

    # Always 8-dim =>> changes based on `StateEncoding`
    state_obs_keys:
        StateEncoding.POS_EULER:    EEF XYZ (3) + Roll-Pitch-Yaw (3) + <PAD> (1) + Gripper Open/Close (1)
        StateEncoding.POS_QUAT:     EEF XYZ (3) + Quaternion (4) + Gripper Open/Close (1)
        StateEncoding.JOINT:        Joint Angles (7, <PAD> if fewer) + Gripper Open/Close (1)

    state_encoding: Type of `StateEncoding`
    action_encoding: Type of action encoding (e.g., EEF Position vs. Joint Position)
"""

from enum import IntEnum

# commented out for now as we don't have this
# from prismatic.vla.datasets.rlds.oxe.utils.droid_utils import zero_action_filter


# Defines Proprioceptive State Encoding Schemes
class StateEncoding(IntEnum):
    # fmt: off
    NONE = -1               # No Proprioceptive State
    POS_EULER = 1           # EEF XYZ (3) + Roll-Pitch-Yaw (3) + <PAD> (1) + Gripper Open/Close (1)
    POS_QUAT = 2            # EEF XYZ (3) + Quaternion (4) + Gripper Open/Close (1)
    JOINT = 3               # Joint Angles (7, <PAD> if fewer) + Gripper Open/Close (1)
    JOINT_BIMANUAL = 4      # Joint Angles (2 x [ Joint Angles (6) + Gripper Open/Close (1) ])
    # fmt: on


# Defines Action Encoding Schemes
class ActionEncoding(IntEnum):
    # fmt: off
    EEF_POS = 1             # EEF Delta XYZ (3) + Roll-Pitch-Yaw (3) + Gripper Open/Close (1)
    JOINT_POS = 2           # Joint Delta Position (7) + Gripper Open/Close (1)
    JOINT_POS_BIMANUAL = 3  # Joint Delta Position (2 x [ Joint Delta Position (6) + Gripper Open/Close (1) ])
    EEF_R6 = 4              # EEF Delta XYZ (3) + R6 (6) + Gripper Open/Close (1)
    # fmt: on


# === Individual Dataset Configs ===
OXE_DATASET_CONFIGS = {
    "rl_bench_v1": {  #
        "image_obs_keys": {"primary": "image", "secondary": "image_1", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "fractal20220817_data": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["base_pose_tool_reached", "gripper_closed"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "kuka": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [
            "clip_function_input/base_pose_tool_reached",
            "gripper_closed",
        ],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bridge": {  # Version of Bridge V2 in Open X-Embodiment mixture
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": None,
        },  # no secondary for now
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bridge_oxe": {  # Version of Bridge V2 in Open X-Embodiment mixture
        "image_obs_keys": {"primary": "image", "secondary": "image_1", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bridge_v2": {  # Version of Bridge V2 I downloaded
        "image_obs_keys": {
            "primary": "image_0",
            "secondary": "image_1",
            "tertiary": "image_2",
            "wrist": None,
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bridge_orig": {  # Original version of Bridge V2 from project website
        "image_obs_keys": {"primary": "image_0", "secondary": "image_1", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bridge_dataset": {  # Original version of Bridge V2 from project website
        "image_obs_keys": {"primary": "image_0", "secondary": "image_1", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "taco_play": {
        "image_obs_keys": {
            "primary": "rgb_static",
            "secondary": None,
            "wrist": "rgb_gripper",
        },
        "depth_obs_keys": {
            "primary": "depth_static",
            "secondary": None,
            "wrist": "depth_gripper",
        },
        "state_obs_keys": ["state_eef", None, "state_gripper"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "jaco_play": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "image_wrist",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state_eef", None, "state_gripper"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_cable_routing": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist45_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["robot_state", None],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "roboturk": {
        "image_obs_keys": {"primary": "front_rgb", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, None],
        "state_encoding": StateEncoding.NONE,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "nyu_door_opening_surprising_effectiveness": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, None],
        "state_encoding": StateEncoding.NONE,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "viola": {
        "image_obs_keys": {
            "primary": "agentview_rgb",
            "secondary": None,
            "wrist": "eye_in_hand_rgb",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["joint_states", "gripper_states"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_autolab_ur5": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "hand_image",
        },
        "depth_obs_keys": {"primary": "depth", "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "toto": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "language_table": {
        "image_obs_keys": {"primary": "rgb", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["effector_translation", None, None, None, None, None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "columbia_cairlab_pusht_real": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["robot_state", None, None, None, None, None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "stanford_kuka_multimodal_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": "depth_image", "secondary": None, "wrist": None},
        "state_obs_keys": ["ee_position", "ee_orientation", None],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "nyu_rot_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "stanford_hydra_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "austin_buds_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "nyu_franka_play_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": "image_additional_view",
            "wrist": None,
        },
        "depth_obs_keys": {
            "primary": "depth",
            "secondary": "depth_additional_view",
            "wrist": None,
        },
        "state_obs_keys": ["eef_state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "maniskill_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {
            "primary": "depth",
            "secondary": None,
            "wrist": "wrist_depth",
        },
        "state_obs_keys": ["tcp_pose", "gripper_state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "furniture_bench_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "cmu_franka_exploration_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "highres_image",
            "secondary": None,
            "wrist": None,
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, None],
        "state_encoding": StateEncoding.NONE,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "ucsd_kitchen_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["joint_state", None],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "ucsd_pick_and_place_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "austin_sailor_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "austin_sirius_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "bc_z": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [
            "present/xyz",
            "present/axis_angle",
            None,
            "present/sensed_close",
        ],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "utokyo_pr2_opening_fridge_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "utokyo_pr2_tabletop_manipulation_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "utokyo_xarm_pick_and_place_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": "image2",
            "wrist": "hand_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["end_effector_pose", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "utokyo_xarm_bimanual_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["pose_r", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "robo_net": {
        "image_obs_keys": {"primary": "image", "secondary": "image1", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_mvp_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "hand_image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["pose", "gripper"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "berkeley_rpt_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "hand_image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["joint_pos", "gripper"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "kaist_nonprehensile_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "stanford_mask_vit_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tokyo_u_lsmo_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "dlr_sara_pour_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "dlr_sara_grid_clamp_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "dlr_edan_shared_control_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "asu_table_top_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "stanford_robocook_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image_1", "secondary": "image_2", "wrist": None},
        "depth_obs_keys": {"primary": "depth_1", "secondary": "depth_2", "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "imperialcollege_sawyer_wrist_cam": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, "state"],
        "state_encoding": StateEncoding.NONE,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "iamlab_cmu_pickup_insert_converted_externally_to_rlds": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["joint_state", "gripper_state"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "uiuc_d3field": {
        "image_obs_keys": {"primary": "image_1", "secondary": "image_2", "wrist": None},
        "depth_obs_keys": {"primary": "depth_1", "secondary": "depth_2", "wrist": None},
        "state_obs_keys": [None, None, None, None, None, None, None, None],
        "state_encoding": StateEncoding.NONE,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "utaustin_mutex": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_fanuc_manipulation": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["joint_state", None, "gripper_state"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "cmu_playing_with_food": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "finger_vision_1",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "cmu_play_fusion": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "cmu_stretch": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["eef_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_gnm_recon": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_gnm_cory_hall": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_gnm_sac_son": {
        "image_obs_keys": {"primary": None, "secondary": None, "wrist": "image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state", None, None],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "droid": {
        "image_obs_keys": {
            "primary": "exterior_image_1_left",
            "secondary": "exterior_image_2_left",
            "wrist": "wrist_image_left",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.EEF_POS,
        "aux_kwargs": {
            "dataset_frame_transform_kwargs": {
                # "chunk_filter_fn": zero_action_filter,
            },
        },
    },
    "fmb": {
        "image_obs_keys": {
            "primary": "image_side_1",
            "secondary": "image_side_2",
            "wrist": "image_wrist_1",
        },
        "depth_obs_keys": {
            "primary": "image_side_1_depth",
            "secondary": "image_side_2_depth",
            "wrist": "image_wrist_1_depth",
        },
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "dobbe": {
        "image_obs_keys": {"primary": "wrist_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "roboset": {
        "image_obs_keys": {
            "primary": "image_left",
            "secondary": "image_right",
            "wrist": "image_wrist",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.JOINT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "rh20t": {
        "image_obs_keys": {
            "primary": "image_front",
            "secondary": "image_side_right",
            "wrist": "image_wrist",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    ### T-DROID datasets
    "tdroid_carrot_in_bowl": {  # "put carrot in bowl" task, 50 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tdroid_pour_corn_in_pot": {  # "pour corn from red bowl into steel pot" task, 50 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tdroid_flip_pot_upright": {  # "flip pot upright" task, 10 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tdroid_move_object_onto_plate": {  # "move <object> onto plate" task, 150 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tdroid_knock_object_over": {  # "knock <object> over" task, 70 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tdroid_cover_object_with_towel": {  # "cover <object> with towel" task, 45 demos @ 5 Hz control
        "image_obs_keys": {"primary": "static_image", "secondary": None, "wrist": None},
        "depth_obs_keys": {
            "primary": "static_depth_image",
            "secondary": None,
            "wrist": None,
        },
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    ### DROID Finetuning datasets
    "droid_wipe": {
        "image_obs_keys": {
            "primary": "exterior_image_2_left",
            "secondary": None,
            "wrist": "wrist_image_left",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["proprio"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    ### LIBERO datasets (modified versions)
    "libero_spatial_no_noops": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "libero_object_no_noops": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "libero_goal_no_noops": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "libero_10_no_noops": {
        "image_obs_keys": {
            "primary": "image",
            "secondary": None,
            "wrist": "wrist_image",
        },
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["EEF_state", None, "gripper_state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    ### New datasets. Only image keys are somewhat correct, ignore rest
    "robo_set": {
        "image_obs_keys": {"primary": "image_right", "secondary": "image_left", "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "aloha_mobile": {
        "image_obs_keys": {"primary": "cam_high", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "tidybot": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "imperialcollege_sawyer_wrist_cam": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": "wrist_image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "kaist_nonprehensile_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_EULER,
        "action_encoding": ActionEncoding.EEF_POS,
    },
    "berkeley_mvp_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "hand_image", "secondary": None, "wrist": "hand_image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "berkeley_rpt_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "hand_image", "secondary": None, "wrist": "hand_image"},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "nyu_rot_dataset_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
    "tokyo_u_lsmo_converted_externally_to_rlds": {
        "image_obs_keys": {"primary": "image", "secondary": None, "wrist": None},
        "depth_obs_keys": {"primary": None, "secondary": None, "wrist": None},
        "state_obs_keys": ["state"],
        "state_encoding": StateEncoding.POS_QUAT,
        "action_encoding": ActionEncoding.JOINT_POS,
    },
}



#!/usr/bin/env python3
"""
Robometer implementation.
Contains the RBM class with three prediction heads for different objectives.

Note: make sure that the forward pass uses all of the
heads or there will be some problems with FSDP sharding.
"""

import torch
import torch.nn as nn
from transformers import PreTrainedModel, Qwen2_5_VLModel

try:
    from transformers import Qwen3VLModel
except ImportError:
    Qwen3VLModel = None

# from transformers import AutoModelForImageTextToText as Molmo2VLModel  # Molmo2 uses AutoModelForImageTextToText
from transformers import SmolVLMModel
import torch.distributed as dist

from roboreason.robometer.robometer.models.utils import ModelOutput
from roboreason.robometer.robometer.models.heads import PredictionHeadsMixin
from roboreason.robometer.robometer.utils.timer import _timer
from roboreason.robometer.robometer.utils.logger import get_logger

logger = get_logger()


def squeeze_last_safe(x: torch.Tensor) -> torch.Tensor:
    if x.ndim > 1 and x.shape[-1] == 1:
        return x.squeeze(-1)
    return x


class RBM(PredictionHeadsMixin, PreTrainedModel):
    """Robometer (RBM) with three prediction heads for different objectives.

    Supports multiple base model architectures:
    - Qwen2.5-VL (Qwen2_5_VLModel)
    - SmolVLM (AutoModelForImageTextToText)
    """

    # unused param i think
    config_class = Qwen2_5_VLModel.config_class

    # Declare support for SDPA and Flash Attention (will delegate to underlying model), needed for Qwen3
    _supports_sdpa = True
    _supports_flash_attn_2 = True

    def __init__(self, config, processor, tokenizer, base_model=None, base_model_id=None, model_config=None):
        if "SmolVLM" in base_model_id:
            hidden_size = config.text_config.hidden_size
            self.model_cls = SmolVLMModel
        elif "Qwen2.5" in base_model_id:
            hidden_size = config.hidden_size
            self.model_cls = Qwen2_5_VLModel
        elif "Qwen3" in base_model_id:
            hidden_size = config.text_config.hidden_size
            self.model_cls = Qwen3VLModel
        elif "Molmo" in base_model_id:
            # Molmo2 is based on Qwen3 architecture
            hidden_size = config.text_config.hidden_size
            self.model_cls = Qwen3VLModel
            # self.model_cls = Molmo2VLModel
        else:
            raise ValueError(f"Unsupported base model: {base_model_id}")

        super().__init__(
            config,
            hidden_dim=hidden_size,
            model_config=model_config,
            dropout=0.1,
        )

        if base_model is not None:
            self.model = base_model
        else:
            self.model = self.model_cls(config)

        self.config_class = self.model_cls.config_class
        self.base_model_id = base_model_id

        self.model_dtype = self.model.dtype
        self.progress_head = self.progress_head.to(dtype=self.model_dtype)
        self.preference_head = self.preference_head.to(dtype=self.model_dtype)
        self.success_head = self.success_head.to(dtype=self.model_dtype)

        self.processor = processor
        self.tokenizer = tokenizer
        self.model_config = model_config

        self.average_temporal_patches = self.model_config.average_temporal_patches
        self.use_per_frame_progress_token = self.model_config.use_per_frame_progress_token
        self.use_multi_image = self.model_config.use_multi_image

        # Frame pooling strategy for multi-image mode (used when NOT using per-frame progress tokens).
        # - mean: average pool patch tokens in the frame span
        # - boundary: use the last patch token in the frame span
        # - attention: learned attention pooling over patch tokens in the frame span
        self.frame_pooling = getattr(self.model_config, "frame_pooling", "mean")
        self.frame_pooling_attn_temperature = float(getattr(self.model_config, "frame_pooling_attn_temperature", 1.0))
        if self.frame_pooling_attn_temperature <= 0:
            raise ValueError("frame_pooling_attn_temperature must be > 0")
        # Always create the attention pooling projection so checkpoints can be loaded across pooling modes.
        self.frame_pool_attn = nn.Linear(hidden_size, 1, bias=False).to(dtype=self.model_dtype)

        # Validate that use_per_frame_progress_token requires use_multi_image
        if self.use_per_frame_progress_token and not self.use_multi_image:
            raise ValueError(
                "use_per_frame_progress_token=True requires use_multi_image=True. "
                "Per-frame progress tokens can only be used in multi-image mode."
            )

        # Molmo2 only supports multi-image mode, not video
        if "Molmo" in self.base_model_id and not self.use_multi_image:
            raise ValueError(
                "Molmo2 does not support video mode (use_multi_image=False). "
                "Please set data.use_multi_image=True to use Molmo2 with multi-image input."
            )

    def gradient_checkpointing_enable(self, **kwargs):
        """Delegates gradient checkpointing enabling to the base model."""
        self.model.gradient_checkpointing_enable(**kwargs)

    def gradient_checkpointing_disable(self, **kwargs):
        """Delegates gradient checkpointing disabling to the base model."""
        self.model.gradient_checkpointing_disable(**kwargs)

    def generate(self, *args, **kwargs):
        """Delegates generation to the base model."""
        return self.model.generate(*args, **kwargs)

    def prepare_inputs_for_generation(self, *args, **kwargs):
        """Delegates input preparation for generation to the base model."""
        return self.model.prepare_inputs_for_generation(*args, **kwargs)

    def _extract_hidden_states_from_token_pairs(
        self,
        hidden_state: torch.Tensor,
        input_ids: torch.Tensor,
    ) -> torch.Tensor:
        """
        Extract image/video frame embeddings from hidden states by finding token pairs and mean pooling.

        This is a general function that works for both SmolVLM and Qwen multi-image mode.
        It automatically detects which model is being used based on the base_model_id:
        - SmolVLM: Uses the same token for start and end: <fake_token_around_image>
        - Qwen: Uses different tokens: <|vision_start|> and <|vision_end|>

        Args:
            hidden_state: Hidden states tensor [seq_len, hidden_dim]
            input_ids: Input token IDs [seq_len]

        Returns:
            frame_embeddings: Tensor [num_frames, hidden_dim] containing mean-pooled
                            embeddings for each frame/image between token pairs
        """
        # Detect model type and get appropriate tokenizer and tokens
        is_molmo = "Molmo" in self.base_model_id
        if "SmolVLM" in self.base_model_id:
            # SmolVLM mode: same token appears in pairs
            tokenizer = self.tokenizer
            start_token = "<fake_token_around_image>"
            end_token = None  # Same token for both start and end
            use_same_token = True
            use_molmo_mode = False
        elif is_molmo:
            # Molmo2 mode: <low_res_im_start> followed by <im_patch> tokens
            tokenizer = self.processor.tokenizer
            start_token = "<low_res_im_start>"
            end_token = None  # No explicit end token
            patch_token = "<im_patch>"
            use_same_token = False
            use_molmo_mode = True
        else:
            # Qwen mode: different start and end tokens
            tokenizer = self.processor.tokenizer
            start_token = "<|vision_start|>"
            end_token = "<|vision_end|>"
            use_same_token = False
            use_molmo_mode = False

        # Get token IDs
        start_token_id = tokenizer.convert_tokens_to_ids(start_token)

        # Find all positions where start tokens appear
        start_positions = (input_ids == start_token_id).nonzero(as_tuple=True)[0]

        if len(start_positions) == 0:
            raise ValueError(
                f"No {start_token} tokens found in input_ids. Token ID {start_token_id} not found in sequence."
            )

        # Handle different pairing modes
        if use_same_token:
            # SmolVLM mode: same token appears in pairs
            if len(start_positions) % 2 != 0:
                raise ValueError(
                    f"Expected even number of {start_token} tokens (pairs), but found {len(start_positions)} tokens."
                )

            # Group tokens into pairs (every two consecutive tokens form a pair)
            token_pairs = []
            for i in range(0, len(start_positions), 2):
                token_pairs.append((start_positions[i].item(), start_positions[i + 1].item()))
        elif use_molmo_mode:
            # Molmo2 mode: <low_res_im_start> followed by <im_patch> tokens
            patch_token_id = tokenizer.convert_tokens_to_ids(patch_token)
            im_patch_positions = (input_ids == patch_token_id).nonzero(as_tuple=True)[0]

            token_pairs = []
            for start_idx, start_pos in enumerate(start_positions):
                start_pos_val = start_pos.item()
                # Find the last consecutive im_patch token after this start
                patches_after_start = im_patch_positions[im_patch_positions > start_pos]
                if len(patches_after_start) > 0:
                    # Find where patches stop (at next image start or end of sequence)
                    if start_idx + 1 < len(start_positions):
                        next_start = start_positions[start_idx + 1].item()
                        patches_for_this_image = patches_after_start[patches_after_start < next_start]
                    else:
                        patches_for_this_image = patches_after_start
                    if len(patches_for_this_image) > 0:
                        end_pos = patches_for_this_image[-1].item()
                        token_pairs.append((start_pos_val, end_pos))
        else:
            # Qwen mode: different start and end tokens
            end_token_id = tokenizer.convert_tokens_to_ids(end_token)

            # Find all positions where end tokens appear
            end_positions = (input_ids == end_token_id).nonzero(as_tuple=True)[0]

            if len(end_positions) == 0:
                raise ValueError(
                    f"No {end_token} tokens found in input_ids. Token ID {end_token_id} not found in sequence."
                )

            if len(start_positions) != len(end_positions):
                raise ValueError(
                    f"Mismatched number of tokens: "
                    f"found {len(start_positions)} {start_token} tokens "
                    f"and {len(end_positions)} {end_token} tokens."
                )

            # Pair up start and end tokens (they should appear in order: start, end, start, end, ...)
            token_pairs = []
            for i in range(len(start_positions)):
                start_pos = start_positions[i].item()
                end_pos = end_positions[i].item()

                if start_pos >= end_pos:
                    raise ValueError(
                        f"Invalid token pair at index {i}: "
                        f"{start_token} at {start_pos}, {end_token} at {end_pos}. "
                        f"Start must come before end."
                    )
                token_pairs.append((start_pos, end_pos))

        # Extract hidden states between token pairs
        frame_embeddings = []
        for start_pos, end_pos in token_pairs:
            # Extract hidden states between the token pair (exclusive of the tokens themselves)
            # Add 1 to start_pos to exclude the start token, end_pos is exclusive
            frame_tokens = hidden_state[start_pos + 1 : end_pos]

            if frame_tokens.shape[0] == 0:
                # If no tokens between the pair, use the token positions themselves
                # This shouldn't happen normally, but handle it gracefully
                frame_embedding = (hidden_state[start_pos] + hidden_state[end_pos]) / 2.0
            else:
                if self.frame_pooling == "mean":
                    frame_embedding = frame_tokens.mean(dim=0)  # [hidden_dim]
                elif self.frame_pooling == "boundary":
                    # Use the last patch token as a boundary summary of the frame span.
                    frame_embedding = frame_tokens[-1]
                elif self.frame_pooling == "attention":
                    # Learned attention pooling over patch tokens in this frame span.
                    # scores: [num_tokens]
                    scores = self.frame_pool_attn(frame_tokens).squeeze(-1) / self.frame_pooling_attn_temperature
                    weights = torch.softmax(scores, dim=0).unsqueeze(-1)  # [num_tokens, 1]
                    frame_embedding = (weights * frame_tokens).sum(dim=0)
                else:
                    raise ValueError(f"Unsupported frame_pooling: {self.frame_pooling}")

            frame_embeddings.append(frame_embedding)

        if len(frame_embeddings) == 0:
            return torch.empty(0, hidden_state.shape[-1], device=hidden_state.device, dtype=hidden_state.dtype)

        return torch.stack(frame_embeddings)  # [num_frames, hidden_dim]

    def _extract_progress_from_trajectory(
        self,
        hidden_state: torch.Tensor,
        start_position: int,
        video_grid_thw: list[int],  # [T, H, W]
        merge_size: int,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Extract progress and success predictions from a trajectory's hidden states.

        Args:
            hidden_state: Hidden states tensor [seq_len, hidden_dim]
            start_position: Starting position in the sequence for this trajectory
            video_grid_thw: Video grid dimensions [T, H, W] where T is number of temporal patch groups,
                           H and W are spatial grid dimensions
            merge_size: Merge size for patch grouping

        Returns:
            tuple: (progress_logits [T], success_logits [T])
        """
        T, H, W = video_grid_thw

        if T == 0:
            return torch.empty(0, device=hidden_state.device), torch.empty(0, device=hidden_state.device)

        # Calculate tokens per frame: (H * W) // merge_size^2
        tokens_per_frame = (H * W) // (merge_size**2)

        if self.average_temporal_patches:
            # Average all tokens within each temporal patch group
            temporal_patch_tokens = []
            current_pos = start_position
            for t_idx in range(T):
                start_idx = current_pos
                end_idx = current_pos + tokens_per_frame
                patch_tokens = hidden_state[start_idx:end_idx]  # [tokens_per_frame, hidden_dim]
                patch_embedding = patch_tokens.mean(dim=0)  # [hidden_dim] - averaged
                temporal_patch_tokens.append(patch_embedding)
                current_pos = end_idx
            boundary_hidden_states = torch.stack(temporal_patch_tokens)  # [T, hidden_dim]
        else:
            # Use last token (boundary) of each temporal patch group
            frame_boundary_positions = []
            current_pos = start_position
            for _frame_idx in range(T):
                frame_end = current_pos + tokens_per_frame
                frame_boundary_positions.append(frame_end)
                current_pos = frame_end

            trajectory_boundaries = torch.tensor(frame_boundary_positions, device=hidden_state.device)
            boundary_hidden_states = hidden_state[trajectory_boundaries]  # [T, hidden_dim]

        assert boundary_hidden_states.shape[0] == T, f"Expected {T} frames, got {boundary_hidden_states.shape[0]}"
        progress_output = self.progress_head(boundary_hidden_states)  # [T, 1] or [T, num_bins] for discrete
        if self.use_discrete_progress:
            progress = progress_output  # [T, num_bins] - keep logits
        else:
            progress = squeeze_last_safe(progress_output)  # [T]
        success = squeeze_last_safe(self.success_head(boundary_hidden_states))  # [T]

        return progress, success

    def _extract_hidden_state_from_token(
        self,
        hidden_state: torch.Tensor,
        input_ids: torch.Tensor,
        token_name: str,
    ) -> torch.Tensor | list[torch.Tensor]:
        """
        Extract hidden states at specific token positions.

        Args:
            hidden_state: Hidden states tensor [B, seq_len, hidden_dim] or [seq_len, hidden_dim]
            input_ids: Input token IDs [B, seq_len] or [seq_len]
            token_name: Name of the token to find (e.g., "<|prog_token|>", "<|pref_token|>")

        Returns:
            If exactly one token per sequence: tensor [B, hidden_dim]
            If multiple tokens per sequence (variable counts): list of tensors, one per batch item [num_tokens_i, hidden_dim]
        """
        # Handle both batched and unbatched inputs
        is_batched = hidden_state.dim() == 3
        if not is_batched:
            hidden_state = hidden_state.unsqueeze(0)  # [1, seq_len, hidden_dim]
            input_ids = input_ids.unsqueeze(0)  # [1, seq_len]

        B = input_ids.shape[0]

        # Get tokenizer (works for both SmolVLM and Qwen)
        if "SmolVLM" in self.base_model_id:
            tokenizer = self.tokenizer
        else:
            tokenizer = self.processor.tokenizer

        # Get token ID
        token_id = tokenizer.convert_tokens_to_ids(token_name)

        # Find all token positions across batch - vectorized
        token_mask = input_ids == token_id  # [B, seq_len]
        batch_indices, positions = token_mask.nonzero(as_tuple=True)  # both [total_tokens]

        if len(positions) == 0:
            raise ValueError(f"{token_name} not found in any sequence")

        # Extract all hidden states at token positions at once - vectorized
        all_hidden = hidden_state[batch_indices, positions]  # [total_tokens, hidden_dim]

        # Count tokens per batch sample
        token_counts = torch.bincount(batch_indices, minlength=B)  # [B]

        # Check if all sequences have exactly one token
        if (token_counts == 1).all():
            # Return as [B, hidden_dim] tensor
            return all_hidden.reshape(B, -1)
        else:
            # Split by batch index into list of tensors (variable token counts)
            return list(torch.split(all_hidden, token_counts.tolist()))

    def _forward_smolvlm(
        self,
        input_ids,
        attention_mask,
        pixel_values,
        sample_type,
        timing_raw,
        **kwargs,
    ):
        """Forward pass for SmolVLM model. Returns (ModelOutput, timing_raw)."""
        model_kwargs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "pixel_values": pixel_values,
            **kwargs,
        }
        with _timer("time/rbm_forward", timing_raw=timing_raw):
            outputs = self.model(**model_kwargs, output_hidden_states=True, return_dict=True)

        hidden_state = outputs.hidden_states[-1]  # [B, seq_len, hidden_dim]
        progress_logits = {"A": None, "B": None}
        success_logits = {"A": None, "B": None}

        # Create output
        output = ModelOutput()

        # Handle token-based extraction first if using per-frame progress tokens
        if self.use_per_frame_progress_token:
            progress_logits, success_logits, pref_or_sim_logits = self._process_token_extraction(
                hidden_state, input_ids, sample_type
            )
            output.progress_logits = progress_logits
            output.success_logits = success_logits
            if pref_or_sim_logits is not None:
                output.pref_logits = pref_or_sim_logits
        else:
            # Process frames normally
            with _timer("time/progress_logits", timing_raw=timing_raw):
                progress_logits, success_logits = self._process_smolvlm_frames(hidden_state, input_ids, sample_type)
            output.progress_logits = progress_logits
            output.success_logits = success_logits

            # Handle preference token if needed
            if sample_type == "preference":
                token_hidden = self._extract_hidden_state_from_token(hidden_state, input_ids, "<|pref_token|>")
                output.pref_logits = self.preference_head(token_hidden)

        return output, timing_raw

    def _process_smolvlm_frames(self, hidden_state, input_ids, sample_type):
        """Process SmolVLM frames and return progress/success logits."""
        B = input_ids.shape[0]
        progress_logits_A, progress_logits_B = [], []
        success_logits_A, success_logits_B = [], []

        for i in range(B):
            frame_embeddings = self._extract_hidden_states_from_token_pairs(hidden_state[i], input_ids[i])
            if frame_embeddings.shape[0] == 0:
                raise ValueError(f"No frame embeddings extracted for sample {i}")

            if sample_type == "progress":
                traj_A, traj_B = frame_embeddings, None
            else:
                mid = frame_embeddings.shape[0] // 2
                traj_A, traj_B = frame_embeddings[:mid], frame_embeddings[mid:]

            # Trajectory A
            prog_A = self.progress_head(traj_A)
            progress_logits_A.append(prog_A if self.use_discrete_progress else squeeze_last_safe(prog_A))
            success_logits_A.append(squeeze_last_safe(self.success_head(traj_A)))

            # Trajectory B
            if traj_B is not None:
                prog_B = self.progress_head(traj_B)
                progress_logits_B.append(prog_B if self.use_discrete_progress else squeeze_last_safe(prog_B))
                success_logits_B.append(squeeze_last_safe(self.success_head(traj_B)))
            else:
                progress_logits_B.append(None)
                success_logits_B.append(None)

        return (
            {
                "A": torch.stack(progress_logits_A) if progress_logits_A else None,
                "B": torch.stack(progress_logits_B) if progress_logits_B and progress_logits_B[0] is not None else None,
            },
            {
                "A": torch.stack(success_logits_A) if success_logits_A else None,
                "B": torch.stack(success_logits_B) if success_logits_B and success_logits_B[0] is not None else None,
            },
        )

    def _forward_qwen(
        self,
        input_ids,
        attention_mask,
        pixel_values,
        pixel_values_videos,
        image_grid_thw,
        video_grid_thw,
        sample_type,
        timing_raw,
        second_per_grid_ts=None,
        **kwargs,
    ):
        """Forward pass for Qwen2.5/Qwen3 models. Returns (ModelOutput, timing_raw)."""
        model_kwargs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "pixel_values": pixel_values,
            "pixel_values_videos": pixel_values_videos,
            "image_grid_thw": image_grid_thw,
            "video_grid_thw": video_grid_thw,
            "second_per_grid_ts": second_per_grid_ts,
            **kwargs,
        }
        with _timer("time/rbm_forward", timing_raw=timing_raw):
            # Qwen3 models may need output_hidden_states=True and use hidden_states instead of last_hidden_state
            is_qwen3 = "Qwen3" in self.base_model_id or (
                hasattr(self.model, "config") and "Qwen3" in str(type(self.model))
            )
            if is_qwen3:
                outputs = self.model(**model_kwargs, output_hidden_states=True, return_dict=True)
                # Qwen3 uses hidden_states tuple, take the last layer
                hidden_state = (
                    outputs.hidden_states[-1] if hasattr(outputs, "hidden_states") else outputs.last_hidden_state
                )
            else:
                outputs = self.model(**model_kwargs)
                hidden_state = outputs.last_hidden_state  # [B, seq_len, hidden_dim]

        progress_logits = {"A": None, "B": None}
        success_logits = {"A": None, "B": None}

        # Create output
        output = ModelOutput()

        if self.use_per_frame_progress_token:
            progress_logits, success_logits, pref_or_sim_logits = self._process_token_extraction(
                hidden_state, input_ids, sample_type
            )
            output.progress_logits = progress_logits
            output.success_logits = success_logits
            if pref_or_sim_logits is not None:
                output.pref_logits = pref_or_sim_logits
        else:
            # Process frames normally
            vision_start_token_id = self.processor.tokenizer.convert_tokens_to_ids("<|vision_start|>")
            vision_end_token_id = self.processor.tokenizer.convert_tokens_to_ids("<|vision_end|>")
            split_token_id = self.processor.tokenizer.convert_tokens_to_ids("<|split_token|>")

            tps = getattr(getattr(self.processor, "video_processor", None), "temporal_patch_size", 2)
            merge_size = getattr(getattr(self.processor, "video_processor", None), "merge_size", 14)

            if self.use_multi_image:
                progress_logits, success_logits = self._process_multi_image_frames(
                    hidden_state,
                    input_ids,
                    sample_type,
                    vision_start_token_id,
                    vision_end_token_id,
                    split_token_id,
                    timing_raw,
                )
            else:
                progress_logits, success_logits = self._process_video_frames(
                    hidden_state,
                    input_ids,
                    video_grid_thw,
                    sample_type,
                    vision_start_token_id,
                    split_token_id,
                    tps,
                    merge_size,
                    timing_raw,
                )

            output.progress_logits = progress_logits
            output.success_logits = success_logits

            # Handle preference token if needed
            if sample_type == "preference":
                token_hidden = self._extract_hidden_state_from_token(hidden_state, input_ids, "<|pref_token|>")
                output.pref_logits = self.preference_head(token_hidden)

        return output, timing_raw

    def _forward_molmo(
        self,
        input_ids,
        attention_mask,
        pixel_values,
        image_grid_thw,
        sample_type,
        timing_raw,
        image_grids=None,
        image_token_pooling=None,
        image_num_crops=None,
        **kwargs,
    ):
        """Forward pass for Molmo2 models (multi-image only). Returns (ModelOutput, timing_raw)."""
        model_kwargs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "pixel_values": pixel_values,
            "image_grid_thw": image_grid_thw,
            "image_grids": image_grids,
            "image_token_pooling": image_token_pooling,
            "image_num_crops": image_num_crops,
            **kwargs,
        }
        with _timer("time/rbm_forward", timing_raw=timing_raw):
            # Qwen3 models may need output_hidden_states=True and use hidden_states instead of last_hidden_state
            is_qwen3 = "Qwen3" in self.base_model_id or (
                hasattr(self.model, "config") and "Qwen3" in str(type(self.model))
            )
            if is_qwen3:
                outputs = self.model(**model_kwargs, output_hidden_states=True, return_dict=True)
                # Qwen3 uses hidden_states tuple, take the last layer
                hidden_state = (
                    outputs.hidden_states[-1] if hasattr(outputs, "hidden_states") else outputs.last_hidden_state
                )
            else:
                outputs = self.model(**model_kwargs)
                hidden_state = outputs.last_hidden_state  # [B, seq_len, hidden_dim]

        progress_logits = {"A": None, "B": None}
        success_logits = {"A": None, "B": None}

        # Create output
        output = ModelOutput()

        # Handle token-based extraction first if using per-frame progress tokens
        if self.use_per_frame_progress_token:
            progress_logits, success_logits, pref_or_sim_logits = self._process_token_extraction(
                hidden_state, input_ids, sample_type
            )
            output.progress_logits = progress_logits
            output.success_logits = success_logits
            if pref_or_sim_logits is not None:
                output.pref_logits = pref_or_sim_logits
        else:
            # Process frames normally
            vision_start_token_id = self.processor.tokenizer.convert_tokens_to_ids("<low_res_im_start>")
            split_token_id = self.processor.tokenizer.convert_tokens_to_ids("<|split_token|>")

            progress_logits, success_logits = self._process_multi_image_frames(
                hidden_state, input_ids, sample_type, vision_start_token_id, None, split_token_id, timing_raw
            )

            output.progress_logits = progress_logits
            output.success_logits = success_logits

            # Handle preference token if needed
            if sample_type == "preference":
                token_hidden = self._extract_hidden_state_from_token(hidden_state, input_ids, "<|pref_token|>")
                output.pref_logits = self.preference_head(token_hidden)

        return output, timing_raw

    def _process_multi_image_frames(
        self,
        hidden_state,
        input_ids,
        sample_type,
        vision_start_token_id,
        vision_end_token_id,
        split_token_id,
        timing_raw,
    ):
        """Process frames in multi-image mode (shared by Qwen and Molmo)."""
        progress_logits_A = []
        progress_logits_B = []
        success_logits_A = []
        success_logits_B = []

        all_trajectory_A_frames = []
        all_trajectory_B_frames = []
        trajectory_A_lengths = []
        trajectory_B_lengths = []
        has_trajectory_B = sample_type != "progress"

        with _timer("time/progress_logits", timing_raw=timing_raw):
            # First pass: extract all frame embeddings
            for i, seq_ids in enumerate(input_ids):
                vision_start_positions = (seq_ids == vision_start_token_id).nonzero(as_tuple=True)[0]
                if len(vision_start_positions) == 0:
                    raise ValueError(f"vision_start_token (id={vision_start_token_id}) not found in sequence {i}")

                # Extract embeddings using _extract_hidden_states_from_token_pairs
                # (handles both Qwen and Molmo token patterns)
                frame_embeddings = self._extract_hidden_states_from_token_pairs(hidden_state[i], seq_ids)

                if frame_embeddings.shape[0] == 0:
                    raise ValueError(f"No frame embeddings extracted for sample {i}")

                # Split into trajectories
                if sample_type == "progress":
                    trajectory_A_frames = frame_embeddings
                    trajectory_B_frames = None
                else:
                    split_positions = (seq_ids == split_token_id).nonzero(as_tuple=True)[0]
                    if len(split_positions) == 0:
                        raise ValueError(f"split_token not found in sequence {i}")
                    split_pos = split_positions[0].item()
                    traj_a_count = sum(1 for pos in vision_start_positions if pos.item() < split_pos)
                    trajectory_A_frames = frame_embeddings[:traj_a_count]
                    trajectory_B_frames = frame_embeddings[traj_a_count:]

                all_trajectory_A_frames.append(trajectory_A_frames)
                trajectory_A_lengths.append(trajectory_A_frames.shape[0])

                if trajectory_B_frames is not None:
                    all_trajectory_B_frames.append(trajectory_B_frames)
                    trajectory_B_lengths.append(trajectory_B_frames.shape[0])
                else:
                    all_trajectory_B_frames.append(None)
                    trajectory_B_lengths.append(0)

            # Batch process trajectory A
            if all_trajectory_A_frames:
                batched_A = torch.cat(all_trajectory_A_frames, dim=0)
                progress_A_out = self.progress_head(batched_A)
                success_A_out = squeeze_last_safe(self.success_head(batched_A))

                if self.use_discrete_progress:
                    progress_A_split = torch.split(progress_A_out, trajectory_A_lengths, dim=0)
                else:
                    progress_A_split = torch.split(squeeze_last_safe(progress_A_out), trajectory_A_lengths, dim=0)
                success_A_split = torch.split(success_A_out, trajectory_A_lengths, dim=0)

                for prog, succ in zip(progress_A_split, success_A_split):
                    progress_logits_A.append(prog)
                    success_logits_A.append(succ)

            # Batch process trajectory B
            if has_trajectory_B:
                valid_B = [(f, l) for f, l in zip(all_trajectory_B_frames, trajectory_B_lengths) if f is not None]
                if valid_B:
                    valid_B_frames, valid_B_lengths = zip(*valid_B)
                    batched_B = torch.cat(valid_B_frames, dim=0)
                    progress_B_out = self.progress_head(batched_B)
                    success_B_out = squeeze_last_safe(self.success_head(batched_B))

                    if self.use_discrete_progress:
                        progress_B_split = torch.split(progress_B_out, list(valid_B_lengths), dim=0)
                    else:
                        progress_B_split = torch.split(squeeze_last_safe(progress_B_out), list(valid_B_lengths), dim=0)
                    success_B_split = torch.split(success_B_out, list(valid_B_lengths), dim=0)

                    valid_idx = 0
                    for frame in all_trajectory_B_frames:
                        if frame is not None:
                            progress_logits_B.append(progress_B_split[valid_idx])
                            success_logits_B.append(success_B_split[valid_idx])
                            valid_idx += 1
                        else:
                            progress_logits_B.append(None)
                            success_logits_B.append(None)
                else:
                    progress_logits_B = [None] * len(all_trajectory_A_frames)
                    success_logits_B = [None] * len(all_trajectory_A_frames)
            else:
                progress_logits_B = [None] * len(all_trajectory_A_frames)
                success_logits_B = [None] * len(all_trajectory_A_frames)

        progress_logits = {
            "A": torch.stack(progress_logits_A) if progress_logits_A else None,
            "B": torch.stack(progress_logits_B) if progress_logits_B and progress_logits_B[0] is not None else None,
        }
        success_logits = {
            "A": torch.stack(success_logits_A) if success_logits_A else None,
            "B": torch.stack(success_logits_B) if success_logits_B and success_logits_B[0] is not None else None,
        }
        return progress_logits, success_logits

    def _process_video_frames(
        self,
        hidden_state,
        input_ids,
        video_grid_thw,
        sample_type,
        vision_start_token_id,
        split_token_id,
        tps,
        merge_size,
        timing_raw,
    ):
        """Process frames in video mode (Qwen only). Returns (progress_logits, success_logits)."""
        progress_logits_A = []
        progress_logits_B = []
        success_logits_A = []
        success_logits_B = []

        with _timer("time/progress_logits", timing_raw=timing_raw):
            for i, seq_ids in enumerate(input_ids):
                vision_start_positions = (seq_ids == vision_start_token_id).nonzero(as_tuple=True)[0]
                if len(vision_start_positions) == 0:
                    raise ValueError(f"vision_start_token not found in sequence {i}")

                if video_grid_thw is None or i >= len(video_grid_thw):
                    raise ValueError(f"video_grid_thw required for video mode")

                # Trajectory A
                grid_idx_A = i if sample_type == "progress" else i * tps
                progress_A, success_A = self._extract_progress_from_trajectory(
                    hidden_state[i], vision_start_positions[0].item(), video_grid_thw[grid_idx_A], merge_size
                )
                progress_logits_A.append(progress_A)
                success_logits_A.append(success_A)

                # Trajectory B (if not progress sample)
                if sample_type != "progress":
                    grid_idx_B = i * tps + 1
                    if grid_idx_B >= len(video_grid_thw):
                        raise ValueError(f"video_grid_thw index {grid_idx_B} out of bounds")
                    progress_B, success_B = self._extract_progress_from_trajectory(
                        hidden_state[i], vision_start_positions[1].item(), video_grid_thw[grid_idx_B], merge_size
                    )
                    progress_logits_B.append(progress_B)
                    success_logits_B.append(success_B)
                else:
                    progress_logits_B.append(None)
                    success_logits_B.append(None)

        progress_logits = {
            "A": torch.stack(progress_logits_A) if progress_logits_A else None,
            "B": torch.stack(progress_logits_B) if progress_logits_B[0] is not None else None,
        }
        success_logits = {
            "A": torch.stack(success_logits_A) if success_logits_A else None,
            "B": torch.stack(success_logits_B) if success_logits_B[0] is not None else None,
        }
        return progress_logits, success_logits

    def _apply_heads_to_hidden_states(
        self, hidden_states_list: list[torch.Tensor]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Apply progress and success heads to a list of hidden states.

        Args:
            hidden_states_list: List of tensors, one per batch item [num_tokens_i, hidden_dim]

        Returns:
            tuple: (progress_logits, success_logits)
                - progress_logits: [B, max_tokens] or [B, max_tokens, num_bins] for discrete (padded)
                - success_logits: [B, max_tokens] (padded)
        """
        progress_list = []
        success_list = []
        for hidden in hidden_states_list:
            if hidden.shape[0] > 0:
                progress_output = self.progress_head(hidden)
                if self.use_discrete_progress:
                    progress_list.append(progress_output)
                else:
                    progress_list.append(squeeze_last_safe(progress_output))
                success_list.append(squeeze_last_safe(self.success_head(hidden)))
            else:
                progress_list.append(torch.empty(0, device=hidden.device))
                success_list.append(torch.empty(0, device=hidden.device))

        progress = torch.stack(progress_list) if progress_list else None
        success = torch.stack(success_list) if success_list else None
        return progress, success

    def _process_token_extraction(
        self,
        hidden_state: torch.Tensor,
        input_ids: torch.Tensor,
        sample_type: str,
    ) -> tuple[dict, dict, torch.Tensor | None]:
        """
        Process token-based extraction for progress and preference predictions.

        Returns:
            tuple: (progress_logits, success_logits, pref_logits)
                - progress_logits: dict with "A" and/or "B" keys
                - success_logits: dict with "A" and/or "B" keys
                - pref_logits: preference logits, or None
        """
        progress_logits = {"A": None, "B": None}
        success_logits = {"A": None, "B": None}
        pref_or_sim_logits = None

        # Extract all <|prog_token|> hidden states (returns list of tensors when multiple tokens)
        all_prog_token_hidden = self._extract_hidden_state_from_token(hidden_state, input_ids, "<|prog_token|>")

        if sample_type == "progress":
            # For progress samples, all tokens belong to trajectory A
            hidden_states_A = all_prog_token_hidden
            progress_logits["A"], success_logits["A"] = self._apply_heads_to_hidden_states(hidden_states_A)
        elif sample_type == "preference":
            # For preference, assume equal number of tokens for A and B
            hidden_states_A = []
            hidden_states_B = []
            for tokens in all_prog_token_hidden:
                num_tokens = tokens.shape[0]
                if num_tokens % 2 != 0:
                    raise ValueError(f"Expected even number of <|prog_token|> tokens, got {num_tokens}")
                mid = num_tokens // 2
                hidden_states_A.append(tokens[:mid])
                hidden_states_B.append(tokens[mid:])

            progress_logits["A"], success_logits["A"] = self._apply_heads_to_hidden_states(hidden_states_A)
            progress_logits["B"], success_logits["B"] = self._apply_heads_to_hidden_states(hidden_states_B)

            token_hidden = self._extract_hidden_state_from_token(hidden_state, input_ids, "<|pref_token|>")
            pref_or_sim_logits = self.preference_head(token_hidden)

        return progress_logits, success_logits, pref_or_sim_logits

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        pixel_values=None,
        pixel_values_videos=None,
        image_grid_thw=None,
        video_grid_thw=None,
        sample_type=None,  # "preference", "progress"
        second_per_grid_ts=None,
        timing_raw=None,
        # Molmo2-specific parameters
        image_grids=None,
        image_token_pooling=None,
        image_num_crops=None,
        video_grids=None,
        video_token_pooling=None,
        **kwargs,
    ):
        """
        Forward pass for the RBM (Robometer).

        Dispatches to model-specific forward methods:
        - SmolVLM: _forward_smolvlm
        - Qwen2.5/Qwen3: _forward_qwen
        - Molmo2: _forward_molmo

        Returns:
            tuple: (ModelOutput, timing_raw dict)
        """
        if timing_raw is None:
            timing_raw = {}

        # Dispatch to model-specific forward
        if "SmolVLM" in self.base_model_id:
            return self._forward_smolvlm(
                input_ids=input_ids,
                attention_mask=attention_mask,
                pixel_values=pixel_values,
                sample_type=sample_type,
                timing_raw=timing_raw,
                **kwargs,
            )
        elif "Molmo" in self.base_model_id:
            return self._forward_molmo(
                input_ids=input_ids,
                attention_mask=attention_mask,
                pixel_values=pixel_values,
                image_grid_thw=image_grid_thw,
                sample_type=sample_type,
                timing_raw=timing_raw,
                image_grids=image_grids,
                image_token_pooling=image_token_pooling,
                image_num_crops=image_num_crops,
                **kwargs,
            )
        else:
            # Qwen2.5 / Qwen3
            return self._forward_qwen(
                input_ids=input_ids,
                attention_mask=attention_mask,
                pixel_values=pixel_values,
                pixel_values_videos=pixel_values_videos,
                image_grid_thw=image_grid_thw,
                video_grid_thw=video_grid_thw,
                sample_type=sample_type,
                timing_raw=timing_raw,
                second_per_grid_ts=second_per_grid_ts,
                **kwargs,
            )



import os
import json
from collections import defaultdict

import cv2
import h5py
import numpy as np
from dataset_upload.helpers import generate_unique_id


DEBUG = False


class Ph2dFrameloader:
    """Pickle-able loader to read frames from an HDF5 file on demand.

    Reads frames from either 'observation.image.right' or 'observation.image.left'.
    Each stored frame is expected to be an encoded image buffer which is decoded
    with cv2.imdecode and converted to RGB.
    """

    def __init__(self, hdf5_path: str, camera: str = "right") -> None:
        self.hdf5_path = hdf5_path
        self.camera = camera.lower()

    def __call__(self) -> np.ndarray:
        key = "observation.image.right" if self.camera == "right" else "observation.image.left"

        with h5py.File(self.hdf5_path, "r") as f:
            if key not in f:
                raise KeyError(f"Key '{key}' not found in {self.hdf5_path}")

            data = f[key]
            encoded_frames = data[()]
            frames = []
            for encoded_frame in encoded_frames:
                frame = cv2.imdecode(encoded_frame, cv2.IMREAD_COLOR)
                frames.append(frame)

        frames_np = np.asarray(frames, dtype=np.uint8)
        if frames_np.ndim != 4 or frames_np.shape[-1] != 3:
            raise ValueError(f"Unexpected frames shape for {self.hdf5_path}: {frames_np.shape} (expected (T,H,W,3))")
        return frames_np


def create_new_trajectory(hdf5_path: str, is_robot: bool, caption: str, camera: str) -> dict:
    trajectory_info = {}
    trajectory_info["id"] = generate_unique_id()
    trajectory_info["task"] = caption
    trajectory_info["frames"] = Ph2dFrameloader(hdf5_path, camera=camera)
    trajectory_info["is_robot"] = is_robot
    trajectory_info["quality_label"] = "successful"
    trajectory_info["partial_success"] = 1
    trajectory_info["data_source"] = "ph2d"
    return trajectory_info


def load_ph2d_dataset(dataset_path: str) -> dict[str, list[dict]]:
    """Load Ph2d dataset organized as folders with HDF5 files and optional metadata.json.

    Expected layout:
        <dataset_path>/
            ph2d_metadata.json   # Optional (dataset-specific). TODO: parse to captions/tasks
            <sequence_1>/
                *.hdf5
            <sequence_2>/
                ...

    Args:
        dataset_path: Root directory containing sequence folders of HDF5 files.

    Returns:
        Mapping: task/caption -> list of trajectory dicts.
    """

    task_data: dict[str, list[dict]] = defaultdict(list)

    # load metadata.json
    all_task_attributes = json.load(open(os.path.join(dataset_path, "ph2d_metadata.json")))["per_task_attributes"]
    for task_name, task_attributes in all_task_attributes.items():
        # load all sequences
        embodiment_type = task_attributes["embodiment_type"]
        if "human" in embodiment_type:
            is_robot = False
        else:
            is_robot = True

        for h5_file in os.listdir(os.path.join(dataset_path, task_name)):
            if not h5_file.endswith(".hdf5"):
                continue

            h5_path = os.path.join(dataset_path, task_name, h5_file)
            loaded_data = h5py.File(h5_path, "r")
            task_description = loaded_data.attrs["description"]

            if "observation.image.right" in loaded_data:
                task_data[task_name].append(create_new_trajectory(h5_path, is_robot, task_description, camera="right"))
            if "observation.image.left" in loaded_data:
                task_data[task_name].append(create_new_trajectory(h5_path, is_robot, task_description, camera="left"))

        if DEBUG:
            break
    print(f"Loaded {sum(len(task_list) for task_list in task_data.values())} trajectories from {len(task_data)} tasks")
    return task_data



#!/usr/bin/env python3
"""
RoboFAC dataset loader for the generic dataset converter for Robometer model training.
Loads MINT-SJTU/RoboFAC-dataset structure: realworld_data/<task>/videos/*.mp4 and simulation_data.

Task descriptions from the dataset card: https://huggingface.co/datasets/MINT-SJTU/RoboFAC-dataset
"""

import json
import re
from pathlib import Path

from dataset_upload.helpers import generate_unique_id, load_sentence_transformer_model
from dataset_upload.video_helpers import load_video_frames
from tqdm import tqdm

# Official task name -> language description (from MINT-SJTU/RoboFAC-dataset dataset card).
# Task names match the card exactly (including typos: UprightStask, PegInsetionSide).
TASK_NAME_TO_DESCRIPTION: dict[str, str] = {
    "SpinStack": "Pick up the cube on the spinning disc and stack it on another cube on the disc.",
    "SpinPullStack": "Pull out the cube on the spinning disc and stack it on another cube on the disc.",
    "MicrowaveTask": "Put the spoon on the table into the cup. Open the door of microwave, put the cup into the microwave and close the door.",
    "SafeTask": "Put the gold bar into the safe, close the door of the safe and rotate the cross knob on the door to lock it.",
    "ToolsTask": "Choose the correct (L-shaped) tools, grasp it to pull the correct (2-pins) charger and plug it.",
    "UprightStask": "Upright the peg and stack it on the cube.",
    "PegInsetionSide": "Insert the peg into the hole on the side of the block.",
    "PullCubeTool": "Grasp the L-shaped tool and pull the cube by it.",
    "PlugCharger": "Grasp the charger and plug it into the receptacle.",
    "InsertCylinder": "Upright the cylinder and insert it into the middle hole on the shelf.",
    "PlaceCube": "Pick up the cube and place it into the box.",
    "LiftPegUpright": "Lift the peg and upright it.",
    "PickCube": "Pick the cube to the target position.",
    "PullCube": "Pull the cube to the red and white target.",
    "PushCube": "Push the cube to the red and white target.",
    "StackCube": "Pick up the cube and stack it on another cube.",
}

# Simulation path task names (e.g. UprightStack-v1) may differ from card names; map to card key.
_SIMULATION_PATH_TO_TASK_KEY: dict[str, str] = {
    "UprightStack": "UprightStask",
    "PegInsertionSide": "PegInsetionSide",
}


class RoboFACFrameLoader:
    """Pickle-able loader that reads RoboFAC video files on demand."""

    def __init__(self, file_path: str):
        self.file_path = file_path

    def __call__(self):
        """Load frames from video file. Returns np.ndarray (T, H, W, 3) uint8."""
        return load_video_frames(Path(self.file_path))


def _snake_to_camel(snake: str) -> str:
    """Convert snake_case to CamelCase (e.g. insert_cylinder -> InsertCylinder)."""
    return "".join(word.capitalize() for word in snake.split("_") if word)


def _realworld_folder_to_task_description(folder_name: str) -> str:
    """Convert realworld folder name to task description using dataset-card mapping.

    E.g. so100_insert_cylinder_error -> InsertCylinder -> full description.
    Falls back to title-cased name if not in TASK_NAME_TO_DESCRIPTION.
    """
    name = folder_name.replace("so100_", "").replace("_error", "").strip("_")
    task_key = _snake_to_camel(name)
    return TASK_NAME_TO_DESCRIPTION.get(task_key) or name.replace("_", " ").strip().title()


def _find_mp4_under(path: Path) -> list[Path]:
    """Find all .mp4 files under path (recursive). Handles videos/ or videos/chunk-000/ etc."""
    if not path.exists():
        return []
    return sorted(path.rglob("*.mp4"))


def _simulation_quality_from_folder(folder_name: str) -> str:
    """Map simulation_data subfolder to quality_label (successful / failure)."""
    name_lower = folder_name.lower()
    if "success" in name_lower and "fail" not in name_lower:
        return "successful"
    if "fail" in name_lower or "error" in name_lower:
        return "failure"
    return "failure"  # default for unknown


def _simulation_path_task_to_description(path_task_name: str) -> str:
    """Map simulation path task name to dataset-card description.

    success_data/ and failure_data/ have one subfolder per task variant, e.g.:
    UprightStack-v1, LiftPegUpright-box, PickCube-apple, MicrowaveTask-fork.
    Strip the trailing -<suffix> to get the base task name for lookup.
    """
    # Strip variant suffix: -v1, -box, -apple, -fork, -gen1, etc.
    base = re.sub(r"-[a-zA-Z0-9_]+$", "", path_task_name)
    task_key = _SIMULATION_PATH_TO_TASK_KEY.get(base, base)
    return TASK_NAME_TO_DESCRIPTION.get(task_key) or path_task_name.replace("-", " ").replace("_", " ").strip().title()


TASK_IDENTIFICATION_KEY = "Task identification"


def _extract_task_identification_desc(annos: dict) -> str:
    """Extract assistant 'value' from annos['Task identification'] conversation only.

    We use only the "Task identification" section (not other anno keys). The value is
    the assistant's reply in that conversation, e.g. "Insert the cylinder into the middle hole of the shelf."
    """
    # Prefer exact key, then case-insensitive match for "Task identification"
    task_id = annos.get(TASK_IDENTIFICATION_KEY)
    if task_id is None:
        for k, v in annos.items():
            if k.strip().lower() == TASK_IDENTIFICATION_KEY.lower():
                task_id = v
                break
    if not isinstance(task_id, list):
        return ""
    for turn in task_id:
        if turn.get("from") == "assistant" and "value" in turn:
            return turn["value"].strip()
    return ""


def _load_test_qa_annos(root: Path) -> tuple[dict[str, str], dict[str, str]]:
    """Load test_qa_realworld and test_qa_sim annos_per_video_split*.json and build video_id -> description.

    Each JSON is a dict: { "video_id": { "video": "path", "task": "InsertCylinder", "annos": { "Task identification": [ { "from": "human", "value": "..." }, { "from": "assistant", "value": "Insert the cylinder into the middle hole of the shelf." } ] } } }.
    We use the "Task identification" assistant value as the language description.
    Tries root/test_qa_realworld, root/test_qa_sim, and root/main/... for each.
    """
    video_id_to_desc: dict[str, str] = {}
    task_folder_to_desc: dict[str, str] = {}
    dirs_to_try = [
        root / "test_qa_realworld",
        root / "test_qa_sim",
        root / "main" / "test_qa_realworld",
        root / "main" / "test_qa_sim",
    ]
    for annos_dir in dirs_to_try:
        if not annos_dir.is_dir():
            continue
        for path in sorted(annos_dir.glob("annos_per_video_split*.json")):
            with open(path, encoding="utf-8") as f:
                data = json.load(f)
            for video_id, entry in data.items():
                if not isinstance(entry, dict):
                    continue
                annos = entry.get("annos", {})
                desc = _extract_task_identification_desc(annos)
                if not desc:
                    continue
                video_id_to_desc[video_id] = desc
                video_path_str = entry.get("video", "")
                if video_path_str:
                    stem = Path(video_path_str).stem
                    video_id_to_desc[stem] = desc
                    task_folder_to_desc[video_path_str.split("/")[0]] = desc
                task = entry.get("task", "")
                if task:
                    task_folder_to_desc[task] = desc
    if video_id_to_desc:
        print(f"  Loaded {len(video_id_to_desc)} video_id->description from test_qa_* annos")
    
    return video_id_to_desc, task_folder_to_desc


# Human prompt that indicates task-identification QA in training_qa.json (we only extract from these).
_TRAINING_QA_TASK_IDENTIFICATION_PROMPT = "identify what task the robot is doing"


def _extract_task_desc_from_training_qa_conversations(convs: list) -> str:
    """Extract language description from training_qa conversations.

    Only use the assistant reply when the human asked the task-identification question
    (e.g. 'Can you identify what task the robot is doing in the provided video?').
    """
    if not isinstance(convs, list):
        return ""
    for i, turn in enumerate(convs):
        if turn.get("from") != "human" or "value" not in turn:
            continue
        human_val = (turn.get("value") or "").lower()
        if _TRAINING_QA_TASK_IDENTIFICATION_PROMPT not in human_val:
            continue
        # Next turn should be assistant with the task description
        if i + 1 < len(convs) and convs[i + 1].get("from") == "assistant" and "value" in convs[i + 1]:
            return (convs[i + 1].get("value") or "").strip()
    return ""


def _load_training_qa(root: Path) -> tuple[dict[str, str], dict[str, str]]:
    """Load training_qa.json and build video_id -> description and task_folder -> description.

    training_qa.json is a list of {
        "id": "09936392-adb5-4f34-9410-7c7305d9c76b",
        "video": "dataset_success_cleaned/MicrowaveTask-fork/stack_error/09936392-adb5-4f34-9410-7c7305d9c76b.mp4",
        "conversations": [
            { "from": "human", "value": "<video>\\nCan you identify what task the robot is doing in the provided video?" },
            { "from": "assistant", "value": "Put the fork in the cup and put them in the microwave" }
        ]
    }.
    We only extract the assistant reply when the human asked the task-identification question
    (contains 'identify what task the robot is doing'). That reply is the language description.
    Tries root/training_qa.json and root/main/training_qa.json.
    """
    video_id_to_desc: dict[str, str] = {}
    task_folder_to_desc: dict[str, str] = {}
    for candidate in (root / "training_qa.json", root / "main" / "training_qa.json"):
        if not candidate.exists():
            continue
        with open(candidate, encoding="utf-8") as f:
            data = json.load(f)
        for item in data:
            video = item.get("video", "")
            video_id = item.get("id") or (Path(video).stem if video else "")
            if not video_id:
                continue
            convs = item.get("conversations", [])
            desc = _extract_task_desc_from_training_qa_conversations(convs)
            if not desc:
                continue
            video_id_to_desc[video_id] = desc
            if video:
                task_folder_to_desc[video.split("/")[0]] = desc
        if video_id_to_desc:
            print(f"  Loaded {len(video_id_to_desc)} video_id->description from {candidate}")
            return video_id_to_desc, task_folder_to_desc
            
    return video_id_to_desc, task_folder_to_desc


def _get_realworld_task_description(
    folder_name: str,
    task_folder_to_desc: dict[str, str] | None,
) -> str:
    """Get language description for realworld folder: prefer training_qa (match by base task name), else TASK_NAME_TO_DESCRIPTION. ipdb if no match."""
    name = folder_name.replace("so100_", "").replace("_error", "").strip("_")
    task_key = _snake_to_camel(name)
    if task_folder_to_desc:
        for qa_key, qa_desc in task_folder_to_desc.items():
            base = re.sub(r"-[a-zA-Z0-9_]+$", "", qa_key)
            card_key = _SIMULATION_PATH_TO_TASK_KEY.get(base, base)
            if card_key == task_key:
                return qa_desc
    task_desc = _realworld_folder_to_task_description(folder_name)
    if task_key not in TASK_NAME_TO_DESCRIPTION:
        import ipdb
        ipdb.set_trace()
    return task_desc


def _get_simulation_task_description(
    path_task_name: str,
    task_folder_to_desc: dict[str, str] | None,
) -> str:
    """Get language description for simulation task: prefer task_folder from training_qa, else TASK_NAME_TO_DESCRIPTION. ipdb if no match."""
    if task_folder_to_desc and path_task_name in task_folder_to_desc:
        return task_folder_to_desc[path_task_name]
    desc = _simulation_path_task_to_description(path_task_name)
    base = re.sub(r"-[a-zA-Z0-9_]+$", "", path_task_name)
    task_key = _SIMULATION_PATH_TO_TASK_KEY.get(base, base)
    if task_key not in TASK_NAME_TO_DESCRIPTION:
        import ipdb
        ipdb.set_trace()
    return desc


def _parse_simulation_video_path(
    video_path: Path,
    root: Path,
    video_id_to_desc: dict[str, str] | None = None,
    task_folder_to_desc: dict[str, str] | None = None,
) -> tuple[str, str, str]:
    """From a video under simulation_data/, derive (task_description, quality_label, data_source).

    Supports two directory layouts:
    - simulation_data/success_data|failure_data/<task_folder>/.../<video>.mp4
      -> quality_folder = success_data|failure_data, path_task_name = task_folder (parts[2])
    - simulation_data/<task_folder>/view0|.../<video>.mp4  (no success_data/failure_data)
      -> path_task_name = task_folder (parts[1]), quality_folder = simulation_data
    """
    try:
        rel = video_path.relative_to(root)
    except ValueError:
        return "Simulation", "failure", "simulation_data"
    parts = rel.parts
    if len(parts) < 3:
        return "Simulation", "failure", "simulation_data"
    if parts[0] != "simulation_data":
        return "Simulation", "failure", "simulation_data"
    # If second component is success_data/failure_data, task folder is parts[2]; else it's parts[1] (e.g. LiftPegUpright-box/view0/...)
    if parts[1] in ("success_data", "failure_data"):
        quality_folder = parts[1]
        path_task_name = parts[2]
    else:
        quality_folder = "simulation_data"
        path_task_name = parts[1]
    quality_label = _simulation_quality_from_folder(quality_folder)
    # Prefer description by video id (stem of .mp4), then by task folder
    video_id = video_path.stem
    if video_id_to_desc and video_id in video_id_to_desc:
        task_desc = video_id_to_desc[video_id]
    else:
        task_desc = _get_simulation_task_description(path_task_name, task_folder_to_desc)
    data_source = f"simulation_data/{quality_folder}"
    return task_desc, quality_label, data_source


def _discover_robofac_trajectories(
    dataset_path: Path,
    *,
    realworld: bool = True,
    simulation: bool = True,
    video_id_to_desc: dict[str, str] | None = None,
    task_folder_to_desc: dict[str, str] | None = None,
) -> list[tuple[Path, str, str, str]]:
    """Discover all video files in RoboFAC dataset structure.

    Expected structure (from MINT-SJTU/RoboFAC-dataset):
        realworld_data/
            so100_insert_cylinder_error/
                videos/
                    *.mp4   OR  videos/chunk-000/*.mp4  (recursive)
            ...
        simulation_data/
            success_data/   or  failure_data/
                <TaskFolder>/   one subfolder per task variant (e.g. UprightStack-v1, LiftPegUpright-box, PickCube-apple)
                    (optional subdirs)
                        *.mp4

    If realworld_data is not found at dataset_path, also tries dataset_path / "main"
    (some download methods put repo content under a main/ subfolder).

    Returns:
        List of (video_path, task, quality_label, data_source) for each trajectory.
    """
    out: list[tuple[Path, str, str, str]] = []

    # Resolve root: support both /path/to/RoboFAC-dataset and /path/to/RoboFAC-dataset/main
    root = dataset_path
    realworld_path = root / "realworld_data"
    if not realworld_path.exists() and (root / "main").is_dir():
        root = root / "main"
        realworld_path = root / "realworld_data"
    if not realworld_path.exists():
        print(f"Warning: realworld_data not found at {dataset_path} or {dataset_path}/main")
    else:
        if realworld:
            for task_dir in sorted(realworld_path.iterdir()):
                if not task_dir.is_dir() or task_dir.name.startswith("."):
                    continue
                task_name = task_dir.name
                videos = _find_mp4_under(task_dir)
                # Realworld videos have generic stems (e.g. episode_000000) that repeat across tasks;
                # use task name only for description, not video_id lookup.
                for vid in videos:
                    task_desc = _get_realworld_task_description(task_name, task_folder_to_desc)
                    out.append((vid, task_desc, "failure", f"realworld_data/{task_name}"))
                if videos:
                    print(f"  realworld_data/{task_name}: {len(videos)} videos")

    if simulation:
        sim_path = root / "simulation_data"
        if sim_path.exists():
            # Discover all mp4s under simulation_data and parse path for task + quality
            videos = _find_mp4_under(sim_path)

            for vid in videos:
                task_desc, quality_label, data_source = _parse_simulation_video_path(
                    vid, root,
                    video_id_to_desc=video_id_to_desc,
                    task_folder_to_desc=task_folder_to_desc,
                )
                out.append((vid, task_desc, quality_label, data_source))
            if videos:
                print(f"  simulation_data: {len(videos)} videos (task/quality from path)")
        else:
            print("Warning: simulation_data not found")

    return out


def load_robofac_dataset(
    dataset_path: str,
    max_trajectories: int | None = None,
    realworld: bool = True,
    simulation: bool = True,
) -> dict[str, list[dict]]:
    """Load RoboFAC dataset and organize by task.

    Args:
        dataset_path: Path to the RoboFAC dataset root (e.g. .../RoboFAC-dataset or .../RoboFAC-dataset/main).
        max_trajectories: Maximum number of trajectories to load (None for all).
        realworld: Include realworld_data subfolders.
        simulation: Include simulation_data.

    Returns:
        Dictionary mapping task names to lists of trajectory dicts (frames, task, quality_label, etc.).
    """
    print("Loading RoboFAC dataset from:", dataset_path)
    dataset_path = Path(dataset_path).resolve()
    if not dataset_path.exists():
        raise FileNotFoundError(f"RoboFAC dataset path not found: {dataset_path}")

    # Build video_id -> description from test_qa_* annos (Task identification) and training_qa.json (human asks "identify what task the robot is doing")
    video_id_to_desc, task_folder_to_desc = _load_training_qa(dataset_path)
    test_vid, test_task = _load_test_qa_annos(dataset_path)
    video_id_to_desc.update(test_vid)
    task_folder_to_desc.update(test_task)

    print("Discovering videos (realworld_data/ and simulation_data/)...")
    traj_list = _discover_robofac_trajectories(
        dataset_path,
        realworld=realworld,
        simulation=simulation,
        video_id_to_desc=video_id_to_desc or None,
        task_folder_to_desc=task_folder_to_desc or None,
    )
    if not traj_list:
        raise FileNotFoundError(
            f"No .mp4 videos found under {dataset_path}. "
            "Check that the path points to the RoboFAC-dataset root (containing realworld_data/ and optionally simulation_data/). "
            "If you downloaded with Hugging Face CLI, the root may be under a 'main' subfolder; the loader will try that automatically."
        )
    if max_trajectories is not None and max_trajectories != -1:
        traj_list = traj_list[:max_trajectories]

    print(f"Found {len(traj_list)} trajectory videos total")

    task_data: dict[str, list[dict]] = {}
    for video_path, task_desc, quality_label, data_source in tqdm(
        traj_list, desc="Building RoboFAC trajectories"
    ):
        frame_loader = RoboFACFrameLoader(str(video_path))
        partial = 1.0 if quality_label == "successful" else 0.0
        trajectory = {
            "frames": frame_loader,
            "actions": None,
            "is_robot": True,
            "task": task_desc,
            "quality_label": quality_label,
            "data_source": data_source,
            "partial_success": partial,
            "id": generate_unique_id(),
        }
        task_data.setdefault(task_desc, []).append(trajectory)

    total = sum(len(v) for v in task_data.values())
    print(f"Loaded {total} trajectories from {len(task_data)} tasks")
    return task_data



#!/usr/bin/env python3
"""
Simple validation script for the Robometer dataset format.
Checks fields and data types only.
"""

import argparse
from typing import Any

import numpy as np

from datasets import Dataset, load_from_disk


def validate_dataset_fields_and_types(dataset: Dataset, sample_size: int = 10) -> dict[str, Any]:
    """Validate dataset fields and data types."""

    print(f"Validating dataset fields and data types on {sample_size} sample entries...")

    validation_results = {
        "valid": True,
        "errors": [],
        "warnings": [],
        "stats": {"dataset_size": len(dataset), "samples_checked": 0},
    }

    # Expected schema for the new format
    expected_fields = [
        "id",
        "task",
        "lang_vector",
        "data_source",
        "frames",
        "is_robot",
        "quality_label",
        "preference_group_id",
        "preference_rank",
    ]

    # Check if dataset has features
    if not hasattr(dataset, "features") or dataset.features is None:
        validation_results["valid"] = False
        validation_results["errors"].append("Dataset has no features defined")
        return validation_results

    print(f"Dataset size: {len(dataset)} entries")
    print(f"Dataset features: {list(dataset.features.keys())}")

    # Check required fields
    for field_name in expected_fields:
        if field_name not in dataset.features:
            validation_results["valid"] = False
            validation_results["errors"].append(f"Missing required field: {field_name}")
        else:
            print(f"✓ Field '{field_name}' present")

    # Sample entries for validation
    sample_indices = np.random.choice(len(dataset), min(sample_size, len(dataset)), replace=False)
    validation_results["stats"]["samples_checked"] = len(sample_indices)

    for idx in sample_indices:
        trajectory = dataset[idx]

        try:
            # Validate each field
            if not isinstance(trajectory["id"], str):
                validation_results["errors"].append(f"Trajectory {idx}: 'id' is not a string")

            if not isinstance(trajectory["task"], str):
                validation_results["errors"].append(f"Trajectory {idx}: 'task' is not a string")

            # lang_vector should be length-384 sequence
            lv = trajectory["lang_vector"]
            if isinstance(lv, np.ndarray):
                if lv.shape != (384,):
                    validation_results["errors"].append(
                        f"Trajectory {idx}: 'lang_vector' shape is {lv.shape}, expected (384,)"
                    )
            elif isinstance(lv, list):
                if len(lv) != 384:
                    validation_results["errors"].append(
                        f"Trajectory {idx}: 'lang_vector' length is {len(lv)}, expected 384"
                    )
                else:
                    # check element types
                    if not all(isinstance(x, (int, float, np.floating, np.integer)) for x in lv):
                        validation_results["warnings"].append(
                            f"Trajectory {idx}: 'lang_vector' contains non-numeric elements"
                        )
            else:
                validation_results["errors"].append(f"Trajectory {idx}: 'lang_vector' has unexpected type {type(lv)}")

            if not isinstance(trajectory["data_source"], str):
                validation_results["errors"].append(f"Trajectory {idx}: 'data_source' is not a string")

            if not isinstance(trajectory["frames"], str):
                validation_results["errors"].append(f"Trajectory {idx}: 'frames' is not a string path")

            if not isinstance(trajectory["is_robot"], bool):
                validation_results["errors"].append(f"Trajectory {idx}: 'is_robot' is not a boolean")

            if not isinstance(trajectory["quality_label"], str):
                validation_results["errors"].append(f"Trajectory {idx}: 'quality_label' is not a string")
            else:
                if trajectory["quality_label"] not in {"successful", "failure", "suboptimal"}:
                    validation_results["warnings"].append(
                        f"Trajectory {idx}: 'quality_label' has unexpected value '{trajectory['quality_label']}'"
                    )

            # preference fields can be None
            if trajectory.get("preference_group_id") is not None and not isinstance(
                trajectory["preference_group_id"], str
            ):
                validation_results["errors"].append(
                    f"Trajectory {idx}: 'preference_group_id' is neither None nor string"
                )
            if trajectory.get("preference_rank") is not None and not isinstance(trajectory["preference_rank"], int):
                validation_results["errors"].append(f"Trajectory {idx}: 'preference_rank' is neither None nor int")

            # Print sample task for first trajectory
            if idx == sample_indices[0]:
                print("\nSample task from first trajectory:")
                print(f"  Task: {trajectory['task']}")
                print(f"  ID: {trajectory['id']}")

        except Exception as e:
            validation_results["errors"].append(f"Trajectory {idx}: Error during validation: {e}")

    if validation_results["errors"]:
        validation_results["valid"] = False

    return validation_results


def print_validation_summary(validation_results: dict[str, Any]):
    """Print validation summary."""

    print("\n" + "=" * 50)
    print("VALIDATION SUMMARY")
    print("=" * 50)

    status = "✅ PASS" if validation_results["valid"] else "❌ FAIL"
    print(f"Status: {status}")

    print(f"Dataset size: {validation_results['stats']['dataset_size']}")
    print(f"Samples checked: {validation_results['stats']['samples_checked']}")

    if validation_results.get("errors"):
        print(f"\nErrors ({len(validation_results['errors'])}):")
        for error in validation_results["errors"][:10]:  # Show first 10 errors
            print(f"  - {error}")
        if len(validation_results["errors"]) > 10:
            print(f"  ... and {len(validation_results['errors']) - 10} more errors")

    if validation_results.get("warnings"):
        print(f"\nWarnings ({len(validation_results['warnings'])}):")
        for warning in validation_results["warnings"][:5]:  # Show first 5 warnings
            print(f"  - {warning}")
        if len(validation_results["warnings"]) > 5:
            print(f"  ... and {len(validation_results['warnings']) - 5} more warnings")

    print("=" * 50)


def main():
    """Main validation function."""

    parser = argparse.ArgumentParser(description="Validate dataset fields and data types")
    parser.add_argument("dataset_path", help="Path to the HuggingFace dataset")
    parser.add_argument("--sample-size", type=int, default=10, help="Number of samples to check")

    args = parser.parse_args()

    # Load dataset
    print(f"Loading dataset from: {args.dataset_path}")
    try:
        dataset = load_from_disk(args.dataset_path)
    except Exception as e:
        print(f"Error loading dataset: {e}")
        return

    print("Dataset loaded successfully.")

    # Run validation
    validation_results = validate_dataset_fields_and_types(dataset, args.sample_size)

    # Print summary
    print_validation_summary(validation_results)

    # Exit with error code if validation failed
    if not validation_results["valid"]:
        exit(1)


if __name__ == "__main__":
    main()




from __future__ import annotations


import argparse
import json
from pathlib import Path
from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch

from roboreason.robometer.robometer.data.dataset_types import ProgressSample, Trajectory

# from roboreason.robometer.robometer.evals.eval_server import compute_batch_outputs
######## this produces the following messages when running:
# Skipping import of cpp extensions due to incompatible torch version 2.9.0+cu128 for torchao version 0.16.0             Please see https://github.com/pytorch/ao/issues/2919 for more info
# /data/sls/scratch/pschro/rr/roboreason/robometer/robometer/utils/setup_utils.py:7: UserWarning: WARNING: Unsloth should be imported before [transformers, peft] to ensure all optimizations are applied. Your code may run slower or encounter memory issues without these optimizations.

# Please restructure your imports with 'import unsloth' at the top of your file.
#   from unsloth import FastVisionModel
# 🦥 Unsloth: Will patch your computer to enable 2x faster free finetuning.
# 🦥 Unsloth Zoo will now patch everything to make training faster!
# 2026-03-16 13:41:22 | INFO     | [Rank 0] roboreason.robometer.robometer.evals.eval_server:<module>:62 | roboreason.robometer.robometer.eval_server logger initialized at level DEBUG

import warnings
import contextlib
import io
warnings.filterwarnings("ignore")
with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
    from roboreason.robometer.robometer.evals.eval_server import compute_batch_outputs


# from roboreason.robometer.robometer.evals.eval_viz_utils import create_combined_progress_success_plot, extract_frames
from roboreason.robometer.robometer.utils.save import load_model_from_hf
from roboreason.robometer.robometer.utils.setup_utils import setup_batch_collator


reward_model = None
tokenizer = None
processor = None
exp_config = None
device = None


def unload_model():
    import gc, torch

    for name in ["reward_model", "tokenizer", "processor", "exp_config"]:
        globals()[name] = None

    gc.collect()
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()
    
# def unload_model():
#     global reward_model, tokenizer, processor, exp_config
#     import gc
#     import torch
#     try: 
#         del reward_model
#     except:
#         pass
#     try:
#         del tokenizer
#     except:
#         pass
#     try:
#         del processor
#     except:
#         pass
#     try:
#         del exp_config
#     except:
#         pass
#     reward_model = None
#     tokenizer = None
#     processor = None
#     exp_config = None
#     gc.collect()
#     torch.cuda.empty_cache()
#     torch.cuda.ipc_collect()
    
def load_model(model_path):
    global reward_model, tokenizer, processor, exp_config
    if reward_model is None:
        # unload_sole()   
        print("Loading Robometer model ...")
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # 
        exp_config, tokenizer, processor, reward_model = load_model_from_hf(
            model_path=model_path,
            device=device,
        )
        # 
        reward_model.eval()

def compute_rewards_per_frame_local(
    model_path: str,
    video_frames: np.ndarray,
    task: str,
    device: Optional[torch.device] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Load RBM from HuggingFace and run inference; return per-frame progress and success arrays."""
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # exp_config, tokenizer, processor, reward_model = load_model_from_hf(
    #     model_path=model_path,
    #     device=device,
    # )

    # reward_model.eval()
    load_model(model_path)   # ensures model is loaded once
    global reward_model, tokenizer, processor, exp_config
    
    batch_collator = setup_batch_collator(processor, tokenizer, exp_config, is_eval=True)
    


    T = int(video_frames.shape[0])
    traj = Trajectory(
        frames=video_frames,
        frames_shape=tuple(video_frames.shape),
        task=task,
        id="0",
        metadata={"subsequence_length": T},
        video_embeddings=None,
    )
    progress_sample = ProgressSample(trajectory=traj, sample_type="progress")
    batch = batch_collator([progress_sample])

    progress_inputs = batch["progress_inputs"]
    for key, value in progress_inputs.items():
        if hasattr(value, "to"):
            progress_inputs[key] = value.to(device)

    loss_config = getattr(exp_config, "loss", None)
    is_discrete = (
        getattr(loss_config, "progress_loss_type", "l2").lower() == "discrete"
        if loss_config else False
    )
    num_bins = (
        getattr(loss_config, "progress_discrete_bins", None)
        or getattr(exp_config.model, "progress_discrete_bins", 10)
    )
    print("Running Robometer inference.")
    results = compute_batch_outputs(
        reward_model,
        tokenizer,
        progress_inputs,
        sample_type="progress",
        is_discrete_mode=is_discrete,
        num_bins=num_bins,
    )

    progress_pred = results.get("progress_pred", [])
    progress_array = (
        np.array(progress_pred[0], dtype=np.float32)
        if progress_pred and len(progress_pred) > 0
        else np.array([], dtype=np.float32)
    )

    outputs_success = results.get("outputs_success", {})
    success_probs = outputs_success.get("success_probs", []) if outputs_success else []
    success_array = (
        np.array(success_probs[0], dtype=np.float32)
        if success_probs and len(success_probs) > 0
        else np.array([], dtype=np.float32)
    )

    return progress_array, success_array



def robometer(frames_final, task_description, model_path=None):
    # def robometer(video_path, task_description):
    # frames = load_frames_input(
    #     # str(args.video),
    #     # fps=float(args.fps),
    #     # max_frames=int(args.max_frames),
    #     video_path,
    #     fps=1000,
    #     max_frames=1000,
    # )
    # # 
    # # frames[0].shape
    # frame_height, frame_width = frames[0].shape[:2]
    # # 
    # frames_final=frames
    # if 'test_videos' in video_path:
    #     if frame_width == 2*frame_height:
    #         # extract left half of each frame since the video has side-by-side views and we only want to use the external view
    #         frames_final=[]
    #         for i in range(len(frames)):
    #             frames_final.append(frames[i][:, :frames[i].shape[1]//2, :])
    
    # frames_final[0].shape
    from roboreason.utils.model_utils import get_model_dir
    if model_path is None:
        model_path = get_model_dir("robometer")
    # 
    rewards, success_probs = compute_rewards_per_frame_local(
        # model_path=args.model_path,
        # model_path='../model_checkpoints/Robometer-4B/',
        model_path=model_path,
        video_frames=np.array(frames_final),
        # task=args.task,
        task=task_description,
    )
    
    rewards = rewards.tolist() if rewards is not None else None
    success_probs = success_probs.tolist() if success_probs is not None else None
    
    # multiply by 100 to convert to percentage
    if success_probs is not None:
        success_probs = [prob * 100 for prob in success_probs]
        rewards = [reward * 100 for reward in rewards]
    # 
    return rewards, success_probs






#!/usr/bin/env python3
from __future__ import annotations

import re
import torch
import io
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Union, Optional, Tuple
from datetime import datetime

import aiohttp
import numpy as np
import requests
import torch

from roboreason.robometer.robometer.data.dataset_types import PreferenceSample, ProgressSample, Trajectory
from roboreason.robometer.robometer.data.datasets.helpers import linspace_subsample_frames, pad_trajectory_to_max_frames_np


def extract_rewards_from_output(outputs: Dict[str, Any]) -> np.ndarray:
    """
    Extract rewards from the output dictionary returned by process_batch_helper.

    Args:
        outputs: Dictionary with 'outputs_preference', 'outputs_progress', 'outputs_similarity'
                 Should contain 'outputs_progress' with 'progress_pred' key

    Returns:
        numpy array of rewards (one per sample)
    """
    if outputs.get("outputs_progress") is None:
        raise ValueError("No progress outputs found in batch outputs")

    outputs_progress = outputs["outputs_progress"]
    progress_pred = outputs_progress.get("progress_pred", [])

    # Extract rewards (last value of each progress prediction)
    # Each progress_list contains progress values for each frame in the subsequence
    rewards_list = []
    for progress_list in progress_pred:
        try:
            if isinstance(progress_list, list) and len(progress_list) > 0:
                # The last value is the reward for this subsequence
                reward = float(progress_list[-1])
                # Clamp to [0, 1] range
                reward = max(0.0, min(1.0, reward))
            else:
                # Default to 0.0 if no valid prediction
                reward = 0.0
        except (ValueError, TypeError, IndexError) as e:
            reward = 0.0
        rewards_list.append(reward)

    return np.array(rewards_list, dtype=np.float32)


def extract_success_probs_from_output(outputs: Dict[str, Any]) -> np.ndarray:
    """
    Extract success probabilities from the output dictionary returned by process_batch_helper.

    Args:
        outputs: Dictionary with 'outputs_success'
                 Should contain 'outputs_success' with 'success_probs' key

    Returns:
        numpy array of success probabilities (one per sample)
    """
    if outputs.get("outputs_success") is None:
        raise ValueError("No success probabilities outputs found in batch outputs")

    outputs_success = outputs["outputs_success"]
    success_preds = outputs_success.get("success_probs", [])
    success_probs = []
    for success_pred in success_preds:
        try:
            if isinstance(success_pred, list) and len(success_pred) > 0:
                # The last value is the success probability for this subsequence
                success_prob = float(success_pred[-1])
                success_probs.append(success_prob)
            else:
                success_probs.append(0.0)
        except (ValueError, TypeError, IndexError) as e:
            success_probs.append(0.0)

    return np.array(success_probs, dtype=np.float32)


def raw_dict_to_sample(
    raw_data: Union[Tuple[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
    max_frames: int = 16,
    sample_type: str = "progress",
) -> Union[ProgressSample, PreferenceSample]:
    """
    Convert raw data dictionary to a ProgressSample or PreferenceSample.

    Args:
        raw_data: Dict with 'frames', 'task', 'id', 'metadata', 'video_embeddings', 'text_embedding' or Tuple of (Dict[str, Any], Dict[str, Any])
        max_frames: Maximum number of frames to use (default: 16)
        sample_type: Either "progress" or "preference" (default: "progress")

    Returns:
        ProgressSample or PreferenceSample
    """

    def _build_trajectory(raw_data: Dict[str, Any], num_frames: int) -> Trajectory:
        processed_item: Dict[str, Any] = {}

        # Process frames
        frames_array = raw_data["frames"]

        # Ensure we have the correct shape: (T, H, W, C)
        if len(frames_array.shape) != 4:
            raise ValueError(f"Expected 4D array (T, H, W, C), got shape {frames_array.shape}")

        # Convert from CxHxW to HxWxC if needed
        if frames_array.shape[1] == 3:
            frames_array = np.transpose(frames_array, (0, 2, 3, 1))

        frames_array, _ = linspace_subsample_frames(frames_array, num_frames)
        dummy_progress = [0.0] * len(frames_array)
        frames_array, _ = pad_trajectory_to_max_frames_np(frames_array, dummy_progress, num_frames, pad_from="right")

        if frames_array.size == 0:
            raise ValueError("No frames processed for example")

        processed_item["frames"] = frames_array
        processed_item["frames_shape"] = frames_array.shape
        processed_item["task"] = raw_data["task"]
        processed_item["lang_vector"] = None
        processed_item["metadata"] = raw_data.get("metadata", None)

        # Process video embeddings using same helper functions
        video_embeddings = raw_data.get("video_embeddings")
        if video_embeddings is not None:
            video_embeddings, _ = linspace_subsample_frames(video_embeddings, num_frames)
            dummy_progress_emb = [0.0] * len(video_embeddings)
            video_embeddings, _ = pad_trajectory_to_max_frames_np(
                video_embeddings, dummy_progress_emb, num_frames, pad_from="right"
            )

        text_embedding = raw_data.get("text_embedding")

        # Convert to tensors if they are numpy arrays
        if video_embeddings is not None and isinstance(video_embeddings, np.ndarray):
            video_embeddings = torch.tensor(video_embeddings)
        if text_embedding is not None and isinstance(text_embedding, np.ndarray):
            text_embedding = torch.tensor(text_embedding)

        processed_item["video_embeddings"] = video_embeddings
        processed_item["text_embedding"] = text_embedding
        processed_item["video_shape"] = video_embeddings.shape if video_embeddings is not None else None
        processed_item["text_shape"] = text_embedding.shape if text_embedding is not None else None

        trajectory = Trajectory(**processed_item)
        return trajectory

    if sample_type == "progress":
        assert isinstance(raw_data, dict), "raw_data must be a dictionary"
        trajectory = _build_trajectory(raw_data=raw_data, num_frames=max_frames)
        return ProgressSample(trajectory=trajectory)
    elif sample_type == "preference":
        assert isinstance(raw_data, tuple), "raw_data must be a tuple"
        assert len(raw_data) == 2, "raw_data must be a tuple of two dictionaries"
        trajectories: List[Trajectory] = []
        for trajectory_data in raw_data:
            trajectory = _build_trajectory(raw_data=trajectory_data, num_frames=max_frames)
            trajectories.append(trajectory)
        return PreferenceSample(chosen_trajectory=trajectories[0], rejected_trajectory=trajectories[1])
    else:
        raise ValueError(f"Unsupported sample_type: {sample_type}")


def build_payload(
    samples: list[PreferenceSample | ProgressSample],
) -> tuple[dict[str, Any], list[dict[str, Any]]]:
    """Build a payload with numpy array handling.

    Args:
        samples: List of samples to convert

    Returns:
        Tuple of (files, sample_data) where:
        - files: Dict of numpy arrays converted to .npy format
        - sample_data: List of sample dictionaries with numpy arrays replaced by file references
    """
    files = {}
    sample_data = []

    for sample_idx, sample in enumerate(samples):
        # Copy the original sample and handle numpy arrays
        processed_sample = sample.model_dump().copy()

        # Handle trajectory objects with numpy arrays
        for key in [
            "chosen_trajectory",
            "rejected_trajectory",
            "trajectory",
        ]:
            if key in processed_sample and isinstance(processed_sample[key], dict):
                trajectory = processed_sample[key]

                # Convert numpy arrays to .npy files
                numpy_fields = ["frames", "lang_vector", "video_embeddings", "text_embedding"]
                for field_name in numpy_fields:
                    # if it is a tensor, first convert it to a numpy array
                    if field_name in trajectory and isinstance(trajectory[field_name], torch.Tensor):
                        trajectory[field_name] = trajectory[field_name].numpy()

                    if field_name in trajectory and isinstance(trajectory[field_name], np.ndarray):
                        # Convert numpy array to .npy file
                        buf = io.BytesIO()
                        np.save(buf, trajectory[field_name])
                        buf.seek(0)
                        file_key = f"sample_{sample_idx}_{key}_{field_name}"
                        files[file_key] = (
                            f"sample_{sample_idx}_{key}_{field_name}.npy",
                            buf,
                            "application/octet-stream",
                        )
                        trajectory[field_name] = {"__numpy_file__": file_key}

        sample_data.append(processed_sample)

    return files, sample_data


def post_batch(url: str, payload: dict[str, Any], timeout_s: float = 120.0) -> dict[str, Any]:
    """POST a batch payload to the evaluation server and return parsed JSON."""
    resp = requests.post(url.rstrip("/") + "/evaluate_batch", json=payload, timeout=timeout_s)
    resp.raise_for_status()
    return resp.json()


def post_batch_npy(
    url: str, files: dict[str, Any], sample_data: list[dict[str, Any]], timeout_s: float = 120.0
) -> dict[str, Any]:
    """POST batch using .npy format for numpy arrays."""
    # Convert sample_data to form data
    data = {f"sample_{i}": json.dumps(sample) for i, sample in enumerate(sample_data)}

    # Send as multipart form data
    resp = requests.post(url.rstrip("/") + "/evaluate_batch_npy", files=files, data=data, timeout=timeout_s)
    resp.raise_for_status()
    return resp.json()


async def post_batch_npy_async(
    session: aiohttp.ClientSession,
    url: str,
    files: dict[str, Any],
    sample_data: list[dict[str, Any]],
    timeout_s: float = 120.0,
) -> dict[str, Any]:
    """Async version of post_batch_npy using aiohttp."""
    # Create FormData for aiohttp
    form_data = aiohttp.FormData()

    # Add files
    for key, (filename, file_obj, content_type) in files.items():
        form_data.add_field(key, file_obj, filename=filename, content_type=content_type)

    # Add sample data
    for i, sample in enumerate(sample_data):
        form_data.add_field(f"sample_{i}", json.dumps(sample))

    headers = {"Connection": "close"}
    # Send as multipart form data using aiohttp
    timeout = aiohttp.ClientTimeout(total=timeout_s)
    async with session.post(
        url.rstrip("/") + "/evaluate_batch_npy", data=form_data, timeout=timeout, headers=headers
    ) as resp:
        resp.raise_for_status()
        return await resp.json()


async def parse_npy_form_data(form_data: Any) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
    """Parse multipart form data to extract numpy arrays and other data.

    Args:
        form_data: FastAPI form data from request.form()

    Returns:
        Tuple of (numpy_arrays dict, other_data dict)
    """
    numpy_arrays = {}
    other_data = {}

    for key, value in form_data.items():
        # Check if this is a file upload (UploadFile object)
        if hasattr(value, "filename") and value.filename:
            # This is a file upload
            if value.filename.endswith(".npy"):
                # Load .npy file (await async read)
                content = await value.read()
                buf = io.BytesIO(content)
                array = np.load(buf)
                numpy_arrays[key] = array
            else:
                # Non-.npy file, skip for now
                continue
        else:
            # This is a string value (form field)
            try:
                # Try to parse as JSON
                other_data[key] = json.loads(value)
            except (json.JSONDecodeError, TypeError):
                # Keep as string if not JSON
                other_data[key] = value

    return numpy_arrays, other_data


def reconstruct_payload_from_npy(
    numpy_arrays: Dict[str, np.ndarray],
    other_data: Dict[str, Any],
    trajectory_keys: Optional[List[str]] = None,
    convert_embeddings_to_torch: bool = False,
) -> List[Dict[str, Any]]:
    """Reconstruct the original payload structure from .npy files and form data.

    The client sends data in this format:
    - Files: sample_0_chosen_trajectory_frames.npy, sample_0_trajectory_frames.npy, etc.
    - Data: sample_0, sample_1, etc. (each containing the full sample JSON with numpy file references)

    Args:
        numpy_arrays: Dictionary of numpy arrays loaded from .npy files
        other_data: Dictionary of other form data
        trajectory_keys: List of trajectory keys to process (default: common keys)
        convert_embeddings_to_torch: Whether to convert embeddings to torch tensors

    Returns:
        List of reconstructed sample dictionaries
    """
    if trajectory_keys is None:
        trajectory_keys = [
            "chosen_trajectory",
            "rejected_trajectory",
            "trajectory",
        ]

    samples = []

    # Process each sample
    for i in range(len(other_data)):
        sample_key = f"sample_{i}"
        if sample_key in other_data:
            # Get the sample data - might already be parsed or might be a string
            sample_data = other_data[sample_key]
            if isinstance(sample_data, str):
                # Parse the sample JSON if it's a string
                sample_data = json.loads(sample_data)

            # Replace numpy file references with actual arrays
            for key, value in sample_data.items():
                if key in trajectory_keys:
                    if isinstance(value, dict):
                        for traj_key, traj_value in value.items():
                            if isinstance(traj_value, dict) and traj_value.get("__numpy_file__"):
                                # Replace with actual numpy array
                                file_key = traj_value["__numpy_file__"]
                                if file_key in numpy_arrays:
                                    value[traj_key] = numpy_arrays[file_key]

                            # Convert embeddings to torch if requested
                            if convert_embeddings_to_torch and traj_key in ["video_embeddings", "text_embedding"]:
                                if traj_key in value and value[traj_key] is not None:
                                    if isinstance(value[traj_key], np.ndarray):
                                        value[traj_key] = torch.tensor(value[traj_key])
                                    elif isinstance(value[traj_key], list):
                                        value[traj_key] = torch.tensor(value[traj_key])

            samples.append(sample_data)

    return samples


def find_video_files(directory: str) -> list[str]:
    """Find all video files in a directory.

    Args:
        directory: Path to directory containing video files

    Returns:
        List of paths to video files
    """
    video_extensions = {".mp4", ".avi", ".mov", ".mkv", ".webm", ".flv", ".wmv", ".m4v"}
    video_files = []

    directory_path = Path(directory)
    if not directory_path.is_dir():
        return []

    for file_path in directory_path.iterdir():
        if file_path.is_file() and file_path.suffix.lower() in video_extensions:
            video_files.append(str(file_path))

    video_files.sort()
    return video_files


def infer_task_from_video_name(video_path: str) -> str:
    """Infer task name from video filename.

    Task is everything before the comma (if comma exists), or everything before success/fail/failure.

    Args:
        video_path: Path to video file

    Returns:
        Inferred task name
    """
    video_name = Path(video_path).stem  # Get filename without extension

    # If there's a comma, task is everything before the comma
    if "," in video_name:
        task_part = video_name.split(",")[0]
    else:
        # Otherwise, split by underscore and remove success/fail/failure suffixes
        parts = video_name.split("_")
        filtered_parts = []
        for part in parts:
            part_lower = part.lower()
            if part_lower not in ["success", "fail", "failure"]:
                filtered_parts.append(part)

        if not filtered_parts:
            return "Complete the task"

        task_part = "_".join(filtered_parts)

    # Split by underscore and join with spaces
    task_words = task_part.split("_")
    task = " ".join(task_words)

    if task:
        # Capitalize first letter of first word, keep rest as is
        task = task[0].upper() + task[1:] if len(task) > 1 else task.upper()
    else:
        task = "Complete the task"

    return task


def setup_output_directory(output_dir: Optional[str], video_path: Optional[str] = None) -> str:
    """Create output directory and return path."""
    if output_dir:
        save_dir = output_dir
    else:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        save_dir = os.path.join(".", f"eval_outputs/{timestamp}")

    os.makedirs(save_dir, exist_ok=True)
    return save_dir
